package prover;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.protobuf.ConcreteCrypto;
import meerkat.protobuf.Crypto;
import org.bouncycastle.math.ec.ECPoint;
import qilin.primitives.concrete.ECGroup;

/**
 * Created by Tzlil on 12/30/2015.
 */
public class ElGamalProofOrganizer {

    private final ECGroup group;
    private final ECPoint g;
    private final ECPoint h;
    private final byte[] gEncoded;
    private final byte[] hEncoded;

    public ElGamalProofOrganizer(ECGroup group, ECPoint g, ECPoint h){
        this.group = group;
        this.g = g;
        this.h = h;
        this.gEncoded = group.encode(g);
        this.hEncoded = group.encode(h);
    }

    public enum OrProofOrder {
        first, second, third, fourth
    }

    public enum TrueCouple {
        left, right, unknown
    }

    //used by prover only
    protected ElGamalProofInput createProofInput(Crypto.RerandomizableEncryptedMessage e1, Crypto.RerandomizableEncryptedMessage e2
            , Crypto.RerandomizableEncryptedMessage e1New, Crypto.RerandomizableEncryptedMessage e2New
            , Crypto.EncryptionRandomness r1, Crypto.EncryptionRandomness r2, boolean switched) throws InvalidProtocolBufferException {

        //convert RerandomizableEncryptedMessage to ElGamalCiphertext
        ConcreteCrypto.ElGamalCiphertext e1ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1);
        ConcreteCrypto.ElGamalCiphertext e2ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2);
        ConcreteCrypto.ElGamalCiphertext e1NewElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1New);
        ConcreteCrypto.ElGamalCiphertext e2NewElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2New);

        return new ElGamalProofInput(e1ElGamal,e2ElGamal,e1NewElGamal,e2NewElGamal,r1,r2,switched);
    }

    // can be used by anyone, e.g Verifier
    public ElGamalProofInput createProofInput(Crypto.RerandomizableEncryptedMessage e1, Crypto.RerandomizableEncryptedMessage e2
            , Crypto.RerandomizableEncryptedMessage e1New, Crypto.RerandomizableEncryptedMessage e2New) throws InvalidProtocolBufferException {

        //convert RerandomizableEncryptedMessage to ElGamalCiphertext
        ConcreteCrypto.ElGamalCiphertext e1ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1);
        ConcreteCrypto.ElGamalCiphertext e2ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2);
        ConcreteCrypto.ElGamalCiphertext e1NewElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e1New);
        ConcreteCrypto.ElGamalCiphertext e2NewElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(e2New);

        return new ElGamalProofInput(e1ElGamal,e2ElGamal,e1NewElGamal,e2NewElGamal);
    }


    public class ElGamalProofInput {


        private final OrProofInput first;
        private final OrProofInput second;
        private final OrProofInput third;
        private final OrProofInput fourth;

        private ECPoint convert2ECPoint(ByteString bs){
            return group.decode(bs.toByteArray());
        }

        private ElGamalProofInput(ConcreteCrypto.ElGamalCiphertext e1, ConcreteCrypto.ElGamalCiphertext e2
                , ConcreteCrypto.ElGamalCiphertext e1New, ConcreteCrypto.ElGamalCiphertext e2New
                , Crypto.EncryptionRandomness r1, Crypto.EncryptionRandomness r2, boolean switched){

            ECPoint e1c1 = convert2ECPoint(e1.getC1());
            ECPoint e1c2 = convert2ECPoint(e1.getC2());
            ECPoint e2c1 = convert2ECPoint(e2.getC1());
            ECPoint e2c2 = convert2ECPoint(e2.getC2());
            ECPoint e1Nc1 = convert2ECPoint(e1New.getC1());
            ECPoint e1Nc2 = convert2ECPoint(e1New.getC2());
            ECPoint e2Nc1 = convert2ECPoint(e2New.getC1());
            ECPoint e2Nc2 = convert2ECPoint(e2New.getC2());

            ECPoint c1_e1NDive1 = group.add(e1Nc1, group.negate(e1c1));
            ECPoint c1_e2NDive1 = group.add(e2Nc1, group.negate(e1c1));
            ECPoint c1_e1NDive2 = group.add(e1Nc1, group.negate(e2c1));
            ECPoint c1_e2NDive2 = group.add(e2Nc1, group.negate(e2c1));

            ECPoint c2_e1NDive1 = group.add(e1Nc2, group.negate(e1c2));
            ECPoint c2_e2NDive1 = group.add(e2Nc2, group.negate(e1c2));
            ECPoint c2_e1NDive2 = group.add(e1Nc2, group.negate(e2c2));
            ECPoint c2_e2NDive2 = group.add(e2Nc2, group.negate(e2c2));

            byte[] c1_e1NDive1Encoded = group.encode(c1_e1NDive1);
            byte[] c1_e2NDive1Encoded = group.encode(c1_e2NDive1);
            byte[] c1_e1NDive2Encoded = group.encode(c1_e1NDive2);
            byte[] c1_e2NDive2Encoded = group.encode(c1_e2NDive2);
            byte[] c2_e1NDive1Encoded = group.encode(c2_e1NDive1);
            byte[] c2_e2NDive1Encoded = group.encode(c2_e2NDive1);
            byte[] c2_e1NDive2Encoded = group.encode(c2_e1NDive2);
            byte[] c2_e2NDive2Encoded = group.encode(c2_e2NDive2);


            if (!switched) {
                this.first = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e1NDive2,c2_e1NDive2
                        ,c1_e1NDive1Encoded,c2_e1NDive1Encoded,c1_e1NDive2Encoded,c2_e1NDive2Encoded
                        ,r1,TrueCouple.left);
                this.second = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e2NDive1,c2_e2NDive1
                        ,c1_e1NDive1Encoded,c2_e1NDive1Encoded,c1_e2NDive1Encoded,c2_e2NDive1Encoded
                        ,r1,TrueCouple.left);
                this.third = new OrProofInput(c1_e1NDive2,c2_e1NDive2,c1_e2NDive2,c2_e2NDive2
                        ,c1_e1NDive2Encoded,c2_e1NDive2Encoded,c1_e2NDive2Encoded,c2_e2NDive2Encoded
                        ,r2,TrueCouple.right);
                this.fourth = new OrProofInput(c1_e2NDive1,c2_e2NDive1,c1_e2NDive2,c2_e2NDive2
                        ,c1_e2NDive1Encoded,c2_e2NDive1Encoded,c1_e2NDive2Encoded,c2_e2NDive2Encoded
                        ,r2,TrueCouple.right);
            } else {
                this.first = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e1NDive2,c2_e1NDive2
                        ,c1_e1NDive1Encoded,c2_e1NDive1Encoded,c1_e1NDive2Encoded,c2_e1NDive2Encoded
                        ,r2,TrueCouple.right);
                this.second = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e2NDive1,c2_e2NDive1
                        ,c1_e1NDive1Encoded,c2_e1NDive1Encoded,c1_e2NDive1Encoded,c2_e2NDive1Encoded
                        ,r1,TrueCouple.right);
                this.third = new OrProofInput(c1_e1NDive2,c2_e1NDive2,c1_e2NDive2,c2_e2NDive2
                        ,c1_e1NDive2Encoded,c2_e1NDive2Encoded,c1_e2NDive2Encoded,c2_e2NDive2Encoded
                        ,r2,TrueCouple.left);
                this.fourth = new OrProofInput(c1_e2NDive1,c2_e2NDive1,c1_e2NDive2,c2_e2NDive2
                        ,c1_e2NDive1Encoded,c2_e2NDive1Encoded,c1_e2NDive2Encoded,c2_e2NDive2Encoded
                        ,r1,TrueCouple.left);
            }

        }


        private ElGamalProofInput(ConcreteCrypto.ElGamalCiphertext e1, ConcreteCrypto.ElGamalCiphertext e2
                , ConcreteCrypto.ElGamalCiphertext e1New, ConcreteCrypto.ElGamalCiphertext e2New){

            ECPoint e1c1 = convert2ECPoint(e1.getC1());
            ECPoint e1c2 = convert2ECPoint(e1.getC2());
            ECPoint e2c1 = convert2ECPoint(e2.getC1());
            ECPoint e2c2 = convert2ECPoint(e2.getC2());
            ECPoint e1Nc1 = convert2ECPoint(e1New.getC1());
            ECPoint e1Nc2 = convert2ECPoint(e1New.getC2());
            ECPoint e2Nc1 = convert2ECPoint(e2New.getC1());
            ECPoint e2Nc2 = convert2ECPoint(e2New.getC2());

            ECPoint c1_e1NDive1 = group.add(e1Nc1, group.negate(e1c1));
            ECPoint c1_e2NDive1 = group.add(e2Nc1, group.negate(e1c1));
            ECPoint c1_e1NDive2 = group.add(e1Nc1, group.negate(e2c1));
            ECPoint c1_e2NDive2 = group.add(e2Nc1, group.negate(e2c1));

            ECPoint c2_e1NDive1 = group.add(e1Nc2, group.negate(e1c2));
            ECPoint c2_e2NDive1 = group.add(e2Nc2, group.negate(e1c2));
            ECPoint c2_e1NDive2 = group.add(e1Nc2, group.negate(e2c2));
            ECPoint c2_e2NDive2 = group.add(e2Nc2, group.negate(e2c2));

            this.first = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e1NDive2,c2_e1NDive2);
            this.second = new OrProofInput(c1_e1NDive1,c2_e1NDive1,c1_e2NDive1,c2_e2NDive1);
            this.third = new OrProofInput(c1_e1NDive2,c2_e1NDive2,c1_e2NDive2,c2_e2NDive2);
            this.fourth = new OrProofInput(c1_e2NDive1,c2_e2NDive1,c1_e2NDive2,c2_e2NDive2);
        }



        public OrProofInput getOrProofInput(OrProofOrder orProofOrder) {
            switch (orProofOrder) {

                case first:
                    return this.first;
                case second:
                    return this.second;
                case third:
                    return this.third;
                case fourth:
                    return this.fourth;
            }
            return null;
        }
    }
    public class OrProofInput{

        public final ECPoint g1;
        public final ECPoint h1;
        public final ECPoint g2;
        public final ECPoint h2;
        public final ECPoint g1Tag;
        public final ECPoint h1Tag;
        public final ECPoint g2Tag;
        public final ECPoint h2Tag;

        protected final byte[] g1Encoded;
        protected final byte[] h1Encoded;
        protected final byte[] g2Encoded;
        protected final byte[] h2Encoded;
        protected final byte[] g1TagEncoded;
        protected final byte[] h1TagEncoded;
        protected final byte[] g2TagEncoded;
        protected final byte[] h2TagEncoded;

        protected final Crypto.EncryptionRandomness x;
        protected final TrueCouple flag;

        private OrProofInput(ECPoint h1, ECPoint h2, ECPoint h1Tag, ECPoint h2Tag
                             ,byte[] h1Encoded, byte[] h2Encoded, byte[] h1TagEncoded, byte[] h2TagEncoded
                , Crypto.EncryptionRandomness x, TrueCouple flag) {
            this.g1 = g;
            this.h1 = h1;
            this.g2 = h;
            this.h2 = h2;
            this.g1Tag = g;
            this.h1Tag = h1Tag;
            this.g2Tag = h;
            this.h2Tag = h2Tag;

            this.g1Encoded = gEncoded;
            this.h1Encoded = h1Encoded;
            this.g2Encoded = hEncoded;
            this.h2Encoded = h2Encoded;
            this.g1TagEncoded = gEncoded;
            this.h1TagEncoded = h1TagEncoded;
            this.g2TagEncoded = hEncoded;
            this.h2TagEncoded = h2TagEncoded;

            this.x = x;
            this.flag = flag;

        }
        private OrProofInput(ECPoint h1, ECPoint h2, ECPoint h1Tag, ECPoint h2Tag) {
            this(h1,h2,h1Tag,h2Tag,null,null,null,null,null,TrueCouple.unknown);
        }
    }
}
